from __future__ import annotations

import sys
from enum import Enum
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from types import TracebackType


class ErrorCode(Enum):
    PARSE_ERROR = 1


class SuccessCode(Enum):
    SUCCESS = 0


type ReturnCode = ErrorCode | SuccessCode


class WatchFsBaseException(Exception):
    code: ErrorCode
    message: str

    def __init__(self, message: str):
        super().__init__(message)
        self.message = message


class ParseError(WatchFsBaseException):
    code = ErrorCode.PARSE_ERROR


def handleUncaughtException(exctype: type[Exception], exception: Exception, trace: TracebackType):
    oldHook(exctype, exception, trace)
    if isinstance(exception, WatchFsBaseException):
        sys.exit(exception.code.value)


sys.excepthook, oldHook = handleUncaughtException, sys.excepthook



from __future__ import annotations

import tomllib
from pathlib import Path

from watchfs import __version__

with Path("pyproject.toml").open("rb") as f:
    project_info = tomllib.load(f)


def test_version():
    assert __version__ == project_info["project"]["version"]



from __future__ import annotations

import asyncio
from functools import wraps
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine


def as_sync[T](async_func: Callable[..., Coroutine[Any, Any, T]]) -> Callable[..., T]:
    @wraps(async_func)
    def sync_func(*args: Any, **kwargs: Any):
        return asyncio.run(async_func(*args, **kwargs))

    return sync_func



from __future__ import annotations

from colored import Back, Fore, Style  # type: ignore


class Badge:
    def __init__(self, name: str, fore: Fore, back: Back) -> None:
        self.name = name
        self.fore = fore
        self.back = back

    def __str__(self) -> str:
        return f"{self.back}{self.fore}{Style.bold} {self.name} {Style.reset}"



from __future__ import annotations

import hashlib
import time
from pathlib import Path

from watchfiles import Change
from watchfiles.filters import BaseFilter


def match_pattern(path: Path, pattern: Path) -> bool:
    if pattern.is_dir():
        return pattern in [path, *path.parents]
    return path.match(str(pattern))


class ExcludeFilter(BaseFilter):
    def __init__(self, cli_arg: str):
        super().__init__()
        self.exclude_patterns = self.parse_cli_arg(cli_arg)

    def __call__(self, change: Change, path: str) -> bool:
        return all(not match_pattern(Path(path), pattern) for pattern in self.exclude_patterns)

    @staticmethod
    def parse_cli_arg(cli_arg: str) -> list[Path]:
        exclude_patterns: list[Path] = []
        for arg in cli_arg.split(","):
            path_patterns = Path(arg).absolute()
            exclude_patterns.append(path_patterns)
        return exclude_patterns


class ChangeCacheFilter(BaseFilter):
    CACHE_MAX_ALIVE_TIME = 1

    def __init__(self):
        super().__init__()
        self.cache: dict[str, tuple[float, str]] = {}

    def __call__(self, change: Change, path: str) -> bool:
        match change:
            case Change.added | Change.modified:
                return not self.lookup_cache(path)
            case Change.deleted:
                return True

    def clean_dead_cache(self) -> None:
        to_delete: list[str] = []
        for path, (timestamp, _) in self.cache.items():
            if time.time() - timestamp > self.CACHE_MAX_ALIVE_TIME:
                to_delete.append(path)
        for path in to_delete:
            del self.cache[path]

    def lookup_cache(self, path: str) -> bool:
        self.clean_dead_cache()
        if path not in self.cache:
            self.cache[path] = (time.time(), self.calc_file_md5(path))
            return False
        _, md5 = self.cache[path]
        if md5 != self.calc_file_md5(path):
            self.cache[path] = (time.time(), self.calc_file_md5(path))
            return False
        return True

    def calc_file_md5(self, path: str) -> str:
        with Path(path).open("rb") as f:
            return hashlib.md5(f.read()).hexdigest()


class CombinedFilter(BaseFilter):
    def __init__(self, filters: list[BaseFilter]):
        self.filters = filters

    def __call__(self, change: Change, path: str) -> bool:
        return all(filter(change, path) for filter in self.filters)


def combine_filters(filters: list[BaseFilter]) -> BaseFilter:
    return CombinedFilter(filters)



from __future__ import annotations

import argparse
import asyncio
import shutil
from pathlib import Path
from typing import TYPE_CHECKING

from aiofiles.os import wrap
from colored import Back, Fore  # type: ignore
from watchfiles import Change, awatch  # type: ignore

from watchfs import __version__
from watchfs.as_sync import as_sync
from watchfs.colorful import Badge
from watchfs.exceptions import ParseError
from watchfs.filters import ChangeCacheFilter, ExcludeFilter, combine_filters
from watchfs.rusty import Err, Ok, Result

if TYPE_CHECKING:
    from watchfiles.filters import BaseFilter

BADGE_ADD = Badge("ADDED", Fore.black, Back.green)
BADGE_DEL = Badge("DELETED", Fore.black, Back.red)
BADGE_MOD = Badge("MODIFIED", Fore.black, Back.blue)
CHANGE_TYPE_TO_BADGE = {
    Change.added: BADGE_ADD,
    Change.deleted: BADGE_DEL,
    Change.modified: BADGE_MOD,
}

copyfile = wrap(shutil.copyfile)


async def handle_added(src_dir: Path, dst_dir: Path, changed: Path):
    await handle_modified(src_dir, dst_dir, changed)


async def handle_modified(src_dir: Path, dst_dir: Path, changed: Path):
    changed_rel_to_src = changed.relative_to(src_dir)
    dst = dst_dir / changed_rel_to_src
    if changed.is_dir():
        for child in changed.iterdir():
            await handle_modified(src_dir, dst_dir, child)
    elif changed.is_file():
        dst.parent.mkdir(parents=True, exist_ok=True)
        await copyfile(changed, dst)


async def handle_removed(src_dir: Path, dst_dir: Path, changed: Path):
    changed_rel_to_src = changed.relative_to(src_dir)
    dst = dst_dir / changed_rel_to_src
    if dst.is_dir():
        for child in dst.iterdir():
            child_rel_to_dst = child.relative_to(dst_dir)
            await handle_removed(src_dir, dst_dir, src_dir / child_rel_to_dst)
        dst.rmdir()
    elif dst.is_file():
        dst.unlink()


async def sync(src_dir: str, dst_dir: str, filter: BaseFilter):
    src = Path(src_dir).absolute()
    dst = Path(dst_dir).absolute()
    async for changes in awatch(src, watch_filter=filter):
        for change, path in changes:
            path = Path(path).absolute()
            print(f"{CHANGE_TYPE_TO_BADGE[change]} {path}")
            match change:
                case Change.added:
                    await handle_added(src, dst, path)
                case Change.modified:
                    await handle_modified(src, dst, path)
                case Change.deleted:
                    await handle_removed(src, dst, path)


def parse_sync_mapping(sync_mapping: str) -> Result[tuple[str, str, bool], ParseError]:
    splited_sync_mapping: list[str] = sync_mapping.split(":")
    bidirectional: bool = False
    if ":" in sync_mapping:
        splited_sync_mapping = sync_mapping.split(":")
    elif "<->" in sync_mapping:
        splited_sync_mapping = sync_mapping.split("<->")
        bidirectional = True
    elif "->" in sync_mapping:
        splited_sync_mapping = sync_mapping.split("->")
    if len(splited_sync_mapping) != 2:
        return Err(ParseError("Invalid sync mapping."))
    src_dir, dst_dir = splited_sync_mapping

    # check if src_dir is a subdirectory of dst_dir
    abs_src_dir = Path(src_dir).resolve()
    abs_dst_dir = Path(dst_dir).resolve()
    if abs_src_dir in abs_dst_dir.parents:
        return Err(ParseError(f"src_dir({abs_src_dir}) is a subdirectory of dst_dir({abs_dst_dir})."))

    return Ok((src_dir, dst_dir, bidirectional))


@as_sync
async def main():
    parser = argparse.ArgumentParser(prog="watchfs", description="Watch and sync files.")
    parser.add_argument("-v", "--version", action="version", version=__version__)
    parser.add_argument("sync_mapping", metavar="SRC_DIR:DST_DIR", type=str, nargs="+", help="Sync mapping file.")
    parser.add_argument("--exclude", type=str, help="Exclude directories or files, separated by comma.")
    parser.add_argument("-cc", "--enable-content-caching", action="store_true", help="Enable content caching.")
    args = parser.parse_args()
    parsed_sync_mapping: list[tuple[str, str]] = []
    for sync_src_with_dst in args.sync_mapping:
        match parse_sync_mapping(sync_src_with_dst):
            case Ok((src_dir, dst_dir, bidirectional)):
                parsed_sync_mapping.append((src_dir, dst_dir))
                if bidirectional:
                    parsed_sync_mapping.append((dst_dir, src_dir))
            case Err(err):
                raise err
    filters: list[BaseFilter] = []
    if args.exclude:
        filters.append(ExcludeFilter(args.exclude))
    if args.enable_content_caching:
        filters.append(ChangeCacheFilter())

    combined_filter = combine_filters(filters)
    print(f"Starting watch {', '.join(f"{src_dst[0]} -> {src_dst[1]}" for src_dst in parsed_sync_mapping)}")
    print("Press Ctrl+C to exit.")
    try:
        await asyncio.gather(*[sync(src_dir, dst_dir, combined_filter) for src_dir, dst_dir in parsed_sync_mapping])
    except asyncio.exceptions.CancelledError:
        print("Bye!")


if __name__ == "__main__":
    main()



from __future__ import annotations


class Some[T]:
    _value: T
    __match_args__ = ("_value",)

    def __init__(self, value: T) -> None:
        self._value = value

    def unwrap(self) -> T:
        return self._value

    def __repr__(self) -> str:
        return f"Some({self._value})"


class Ok[T]:
    _value: T
    __match_args__ = ("_value",)

    def __init__(self, value: T) -> None:
        self._value = value

    def ok(self) -> T:
        return self._value

    def __repr__(self) -> str:
        return f"Ok({self._value})"


class Err[E: Exception]:
    _e: E
    __match_args__ = ("_e",)

    def __init__(self, e: E) -> None:
        self._e = e

    def err(self) -> E:
        return self._e

    def __repr__(self) -> str:
        return f"Err({self._e})"


type Result[T, E: Exception] = Ok[T] | Err[E]
type Option[T] = Some[T] | None
